#!/bin/perl5.8

=head1 Name

Dredu --- General Module for Data Reduction

	Created: 2007-09-05
	Updated: 2007-09-06
	Updated: 2007-12-24
	Updated: 2008-01-01 Function dummy()
	Updated: 2008-04-29 Function Max() and Min(); Function miss() modified
	Updated: 2008-05-06 Function Recode(), Dec(), Dec89()
	Updated: 2011-01-31 Recode() modified. New functions Multiply(), Interact().
	Updated: 2011-11-09 Method copy()
	Updated: 2011-11-09 New function Test_hash(). New method input().

=head1 Usage

use Dredu

=cut

package Dredu;

use Exporter ();
@ISA = qw(Exporter);
@EXPORT = qw( %missing_value %missing_type miss valid All_valid miss_miss Missval Misstype misstype Common_question Interact Multiply Sum Max Min dummy Recode Dec Dec89 Test_hash); 
@EXPORT_OK = qw( show_count );


# ************************************
#  Declare Constants and Gloval variables
# ************************************

=head1 Global Variables

=head2 Record separater for input

# $/ = "\n";

=cut

# $/ = "\n";

=head2 Record separater for output

# $\ = "\n";

=cut

# $\ = "\n";

=head2 Field separater for output

# $, = "\t";

=cut

# $, = "\t";


=head1 Exported Variables

=head2 %missing_value


=cut

%missing_value = ();
%missing_type = ();

# ************************************
# Methods and Subroutines
# ************************************

=head1 Exported Subroutines

=head2 miss(@vars) - to return list of missing values.

=cut

sub miss {
	map{ exists $missing_type{$_} ? $_ : () }  @_ ;
}

=head2 valid(@vars) - to return list of non-missing values.

=cut

sub valid {
	map{ exists $missing_type{$_} ? () : $_ }  @_ ;
}

=head2 All_valid(@vars) - to return 1 if @vars include no missing value

=cut

sub All_valid {
	foreach(@_){
		return 0  if exists $missing_type{$_};
	}
	return 1  if @_;
	warn( "No argument for All_valid\n" );
	0;
}

=head2 Recode( $x, %map ) - Convert value according to the map
		Argument: $x = the value to be converted
			%map = the hash according to which the value should be converted.
			$x and keys for %map can be an array reference as well as a scalar.
		Return value: The converted value, if the value is on the map. 
			Otherwise the original value will be returned 

=cut

sub Recode {
	my ( $x, %map ) = @_;
	my @x=();
	if( ref $x ){ @x = @$x ; }
	else { @x = ($x) ; } 
	
	my %m = %map;
	my ($k, $v , $a);
	while( ($k, $v ) = each %m ) {
		if( ref $k ) {
			foreach $a ( @$k ){
				warn( "Already existing key: $a => $v" )  if exists $map{$a};
				$map{$a} = $v;
			}
		}
	}
	my @ret =();
	foreach $x(@x) {
		if( exists $map{$x} ) {
			push @ret, $map{$x} ;
		}
		else {
			push @ret, $x ;
		}
	}
	
	return $ret[0]  if 1==@ret;
	@ret;
}

=head2 Test_hash( $name, $hid, $ext ) - Return hash value for given id
		Hash entries will be read from the file "$name$ext" for the first time.
		Argument: $name = the name for the hash
			$hid = hash key
			$ext (optional) = extention for the input file 
				(default = $_HASH_EXT)
		Return value: The value converted by the hash

=cut

$_HASH_EXT = '.dat' ;
sub Test_hash {
	my ( $name, $hid, $ext ) = @_;
	unless( %{ "$name" } ){
		$ext ||= $_HASH_EXT ;
		open( FILE, $name.$ext ) || die( "Cannot open file $name$ext" );
		local($/) = "\n" ;
		while(<FILE>) {
			my( $key, $val ) = split;
			${ "$name" }{$key} = $val;
		}
		close FILE;
	}
	${ "$name" }{$hid};
}


=head1 Methods

=head2 new($handle) - Construct an object for "Dredu" class

	$handle: Identifier used for the input filehandle.
		If ommitted, the default value 'INFILE' will be used instead.

=cut

sub new {
	my ($class, $handle) = @_;
	my $type = ref($class) || $class;
	$handle ||= 'INFILE' ;	# default
	my $self = {
		handle => $handle,	# file handle
		configfile => '' ,	# configuration file
		datafile => '' ,	# data file to be read
#		rs => $/ ,       	# record separater for input
		fs => $, ,       	# field separater for input
		template => '' ,	# template to unpack data input from a binary file
		para => {} ,    	# hash for parameters for configration
		miss => {} ,    	# hash for missing values
		var => {} ,     	# hash for variables
		variables => [] ,	# list of variables
	}
	;
	bless $self, $type;
	return $self;
}

=head2 copy() - Construct an object as a copy of the given object

=cut

sub copy {
	my ($class) = @_;
	die ( "copy() is an instance method, not for $class.")  unless ref($class);
	my $self = $class->new ;
	%$self = %$class ;
	$self -> set_variables ($self->variables);
	return $self;
}


=head2 missing(%hash) - define/return missing values

	Argument: hash ( name => value )

	Result: the hash will be set to %missing_value and be exported.
		If the method is indirectly called via an object,
		the same hash will be set to $obj->{miss}

This method is called internally when initializing the object.

=cut

sub missing {
	my ( $self, %miss ) = @_;
	if( %miss ){
		%missing_value = %miss;
		%missing_type = reverse %missing_value;
		%{ $self ->{miss} } = %miss if ref($self);
	}
	%missing_value;
}

=head2 Missval - return the missing value corresponding to the missing type

	Return: a value defined in %missing_value

=cut

sub Missval {
	return $missing_value{$_[0]} if exists $missing_value{$_[0]} ;
	warn ( "Not a missing type: $_[0]: " );

	undef;
}

=head2 Misstype - return the type of the missing value

	Return: a string defined in %missing_value

=cut

sub Misstype {
	return $missing_type{$_[0]} if exists $missing_type{$_[0]} ;
	warn ( "Not a missing value: $_[0]\n" );
	undef;
}

=head2 misstype(@v) - Convert missing values to the missing types

	Return: An array of valid values and missing-type strings defined in %missing_value
	
	The returned array includes valid values. 
	If you want to get missing values only, write misstype(miss(@v))

=cut

sub misstype {
	map { miss($_) ? $missing_type{$_} : $_ } @_ ;
}

sub miss_miss {
	my @ret;
	foreach(values %missing_value){
		push @ret, $_, $_;
	}
	@ret;
}

=head2 para(%hash) - set/return parameters

	Argument: hash ( name => value )

This method is called internally when initializing the object.

=cut

sub para {
	my ( $self, %arg ) = @_;
	die ( "para() is an instance method, not for $self.")  unless ref($self);
	if( %arg ){
		%{ $self ->{para} } = %arg;
	}
	%{ $self ->{para} };
}

=head2 variables(@vars) - define/return the variables

	Argument: list of variable names

=cut

sub  variables {
	my ( $self, @arg ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	if( @arg ){
		$self->{variables} = \@arg;
	}
	@{ $self->{variables} };
}

=head2 alias($org, $alias, $alias2, ...  ) - define aliases for a variable

	Argument: the variable name, and array of aliases

=cut

sub  alias {
	my ( $self, $org, @a ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	die ( "No original name.")  unless $org;
	die ( "No alias name.")  unless @a;
	foreach(@a) {
		$self->{var}{$_} = $self->{var}{$org} ;
	}
	return scalar(@a);
}

=head2 set_variables(@vars) - set variables and reference for the variables

	Argument: list of variable names

=cut

sub  set_variables {
	my ( $self, @arg ) = @_;
	die ( "set_variables() is an instance method, not for the $self class.")  unless ref($self);
	if( @arg ){
		$self->{variables} = \@arg;
	}

	if( @{ $self->{variables} } ){
		my( $v, $vname );
		foreach $v ( @{ $self->{variables} } ) {
			$vname = "main::$v";
			if( $v=~/^[A-Z][A-Z_0-9]*$/ && ! defined $$vname) {
				$self->{var}{$v} = \${$vname};
			}
			elsif( $v=~/^\w+$/ ){
				$vname = "Dredu::Var::$v";
				$self->{var}{$v} = \${$vname};
			}
			else {
				my $r;
				$self->{var}{$v} = \$r;
			}
		}
	}
	else {
		die ( "No variables list.") ;
	}

	@{ $self->{variables} };
}


=head2 datafile($fname,$num) - define the data file

	Argument: 

=cut

sub datafile {
	my ( $self, $fname, $num ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);

	my $handle = $self->{handle};
	open( $handle, $fname ) || die("Cannot open file $fname") ;
	
	my $ret='' ;
	while( $num-- ){
		$ret .= <$handle>;
	}
	$ret;
}

=head2 configfile($fname) - define the configration file

	Argument: 

=cut

sub configfile {
	my ( $self, @arg ) = @_;
	die ( "configfile() is an instance method, not for the $self class.")  unless ref($self);
	unless( @arg ){
		@arg = @ARGV;
	}
	
	my $f;
	foreach $f (@arg){
		open( CONFIG , $f ) || die( "Cannot open configration file $f\n" );
		
			#	Read Configuration File
		my( $block, @var, %para, %missing, $file, $key, $val, %alias );
		while(<CONFIG>){
			next if /^\s*$/;
			( /^\[(\w+)\]/ ) && ( $block=$1 ) && next;
			if( $block eq 'var' ){
				chomp;
				push @var, (split(/\t/))[0];
			}elsif( $block eq 'output' ){
				chomp;
				($key, $val) = split(/\t/) ;
				$para{$key} = $val;
			}elsif( $block eq 'miss' ){
				chomp;
				($key, $val) = split(/\t/) ;
				$missing{$key} = $val;
			}elsif( $block eq 'alias' ){
				chomp;
				($key, $val) = split(/\t/) ;
				$alias{$key}=$val;
			}
		}
		close(CONFIG);

		$self -> para(%para);
		$self -> missing(%missing);
		$self -> variables(@var);
		$self -> datafile($para{filename}) if $para{filename};

		$self -> set_variables(@var);
		while( ($key,$val) = each %alias ){
			$self -> alias($key, $val);
		}
	}
}


=head2 values(@name) - refer the value of the variables

	Argument: 

=cut

sub values {
	my ( $self, @arg ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	my @ret=();
	@arg = @{ $self->{variables} }  unless @arg;

	foreach(@arg){
		push @ret, ${ $self->{var}{$_} };
	}
	@ret;
}

=head2 Value($name,$val) - refer or assign the value of the variable

	Argument: 

=cut

sub Value {
	my ( $self, $name, $val ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	unless( $name ){
		die ( "No argument for Value.");
	}
	elsif($val) {
		${ $self->{var}{$name} } = $val;
	}
	${ $self->{var}{$name} }
}


=head2 template($t) - define template for unpack

	Argument: 

=cut

sub template {
	my ( $self, @arg ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	if( @arg ){
	}
}


=head2 fs($r) - define field separater for input

	Argument: 

=cut

sub fs {
	my ( $self, $fs ) = @_;
	die ( "This is an instance method, not for $self.")  unless ref($self);
	if( $fs ){
		$self->{fs}  = $fs;
	}
	$self->{fs} ;
}


=head2 binary_file - switch to binary mode

	Argument: 

=cut

sub binary_file {
	my ( $self ) = @_;
	die ( "read() is an instance method, not for $self.")  unless ref($self);
	binmode $self->{handle};
}

=head2 read_binary - read one record from a binary data file

	Argument: 

=cut

sub read_binary {
	my ( $self ) = @_;
	die ( "read_binary() is an instance method, not for $self.")  unless ref($self);
	binmode( $self->{handle} );

	my $record;
	my $ret = read( $self->{handle}, $record , $self->{para}{recordsize} ) ;
	return $ret unless $ret;
	
	my @v = unpack( $self->{para}{template}, $record );

	my $name = $self->{variables} ;

	my $i;
	my $n = scalar( @$name ) -1;
	
	foreach $i( 0..$n ){
		${ $self->{var}{${$name}[$i]} } = $v[$i];

# print STDERR "[$i]" , ${$name}[$i];

	}
	$ret;
}

=head2 read_text - read one record from the text data file

	Argument: 

=cut

sub read_text {
	my ( $self, $ncard ) = @_;
	die ( "read_text() is an instance method, not for $self.")  unless ref($self);

	$ncard ||= 1;
	my $h = $self->{handle};
	my $record = <$h>;	 # for( 1..$ncard )
	return 0  unless $record;

	my @v = split( $self->fs, $record );

	my $name = $self->{variables} ;

	my $i;
	my $n = scalar( @$name ) -1;
	foreach $i( 0..$n ){
		${ $self->{var}{${$name}[$i]} } = $v[$i];

	}
	$ncard;
}

=head2 input(@s) - read one record from a string

	Argument: @s: array of fields for the record

=cut

sub input {
	my ( $self , @v ) = @_ ;
	die ( "input(s) is an instance method, not for $self.")  unless ref($self);

	my $name = $self->{variables} ;

	my $i;
	my $n = scalar( @$name ) -1;
	foreach $i( 0..$n ){
		${ $self->{var}{${$name}[$i]} } = $v[$i];
	}
	$n;
}



=head2 show_count() - print the count number to STDERR 

	Argument: none

show_count() can be exported explicitly.
	
=cut

my $_n;
sub show_count {
	printf STDERR "\r%d: " , ++$_n;
}

=head2 Common_question() - Extract one value from plural questions for sub-questionnares

	Argument: list of values

Common_question() will be exported by default.
	
=cut

sub Common_question {
	my @val = @_;
	my @ret = valid( @val );
	warn( "More than one valid values for a common question: @ret\n" )  if 1<@ret;
	@ret = grep { $_ != Missval('OUT') } @val  unless @ret;
	$ret[0] = Missval('NA')  unless @ret;
	$ret[0];
}


=head2 Dec() - Decimal vector 

	Argument: list of values (ex.: 0, 3, 5, 1, 2)
	Return value: A scalar of decimal vactor (ex: 3512).
		Missing if the argument includes a missing value.
	
=cut

sub Dec {
	my @v = reverse @_;
	my @m = miss(@v);
	return $m[0] if @m;

	my $ret;
	my $d=10;
	my $p=1;
	foreach(@v){
		$ret += $p*$_;
		$p *= $d;
	}
	$ret;
}

=head2 Dec89() - Decimal vector (call Dec() after converting missing values)

	Argument: list of values (ex.: 0, 3, 5, 1, 2)
	Return value: A scalar of decimal vactor (ex: 3512).
		Missing values are converted as ( NA=>9, OUT=>8 )
	
=cut

sub Dec89 {
	my @v =  map {Recode($_, Missval('OUT')=>8, Missval('NA')=>9)} @_ ;
	Dec(@v);
}

=head2 Sum() - Summation without missing values

	Argument: list of values

Sum() will be exported by default.
	
=cut

sub Sum {
	my @v = valid(@_);
	my $ret=0;
	foreach(@v){
		$ret += $_;
	}
	$ret;
}

=head2 Multiply() - Multiply without missing values

	Argument: list of values

Multiply() will be exported by default.
	
=cut

sub Multiply {
	my @v = valid(@_);
	return '' unless @ret;
	my $ret=1;
	foreach(@v){
		$ret *= $_;
	}
	$ret;
}

=head2 Interact() - Multiply if all values are not a missing value

	Argument: list of values

Interact() will be exported by default.

=cut

sub Interact {
	my (@m) = miss(@_);
	return $m[0] if @m;
	Multiply( valid(@_) ) ;
}

=head2 Max() - The maximum value among the arguments, unless they have no missing value

	Argument: list of values

Max() will be exported by default.
	
=cut

sub Max {
	return (miss(@_))[0]  if miss(@_);
	my ($a) = sort {$b<=>$a} @_ ;
	$a;
}

=head2 Min() - The minimum value among the arguments, unless they have no missing value

	Argument: list of values

Min() will be exported by default.
	
=cut

sub Min {
	return (miss(@_))[0]  if miss(@_);
	my ($a) = sort {$a<=>$b} @_ ;
	$a;
}


=head2 dummy( $v, @list_of_values ) - set dummy variables

	Argument: variable and list of values

	Return value: list of dummy variables
		if $v is a missing value, list of the same missing values

=cut


sub dummy {
	my( $v, @values ) = @_;
	my @ret = ();
	if(miss($v)){
		foreach( 0 .. $#values){
			$ret[$_] = $v;
		}
		return @ret;
	}
	else {
		foreach( 0 .. $#values){
			$ret[$_] = ( $values[$_]==$v ) ? 1 : 0;
		}
		return @ret;
	}
	();
}



# ********************************
1;	# end
# ********************************


=head1 Author

=over 4

=item B<TANAKA Sigeto> <tsigeto@nik.sal.tohoku.ac.jp>

Graduate School of Arts and Letters, Tohoku University

Kawauti 27-1, Aoba Ku, Sendai, Miyagi, 980-8576 JAPAN

WWW: http://www.sal.tohoku.ac.jp/~tsigeto/

=back

=head1 Copyright

	Copyright (c) 2007 TANAKA Sigeto

You may use, duplicate, revise, or redistribute 
this script, without permission from the copyright holder.

=cut
